import common ;
import tests ; 
import cell ; 
import rectcell ;

/**************************************
 *                                    * 
 *               Boxes                * 
 *                                    * 
 **************************************/

int boxfakelinks = 8 ;

int outer (int n) { return boxfakelinks+2n ; }
int inner (int n) { return boxfakelinks+2n+1 ; }
int fouter (int n) { return boxfakelinks-1-2(n%4) ; }
int finner (int n) { return boxfakelinks-1-2(n%4)-1 ; }
int boxnumports (cell c) { return (int) ((c.port.length-boxfakelinks)/2) ; }


real before(real r) { return r-realEpsilon ; }
real after(real r) { return r+realEpsilon ; }
real unint(real r) { return (real)(int)r == r ? after(r) : r ; }

real normalize(real x) { return x>1 || x<0 ? x%1 : x ; }

cell box (real main, Label L="!", real radius, rectangle rect) {

	cell c ; 

	pair E=unit(rect.B-rect.A) ;
	pair N=rotate(90)*E ; 
	pair W=rotate(90)*N ; 
	pair S=rotate(90)*W ; 

	pair P = point(rect.base(0),normalize(main)) ;
	pair Pr = P+radius*E ;
	pair Pu = P+rotate(90)*(Pr-P) ;
	pair Pl = P+radius*W ;
	pair Pd = P+rotate(90)*(Pl-P) ;

	draw(c.pic,Pr--rect.B--rect.C--rect.D--rect.A--Pl) ; 

	path circle = Pr..Pu..Pl..Pd..cycle ;

	draw(c.pic,circle) ;
	label(c.pic,ensuremath(L),position=P) ;
	
	c.port[outer(0)]=port(Pd,Pd-P,defaultportlength-radius) ;
	c.port[inner(0)]=port(Pu,Pu-P,defaultportlength-radius,forwards=false) ;

	for (int i=0;i<4;++i) {
		path b=rect.base(i) ;
		c.port[fouter(i)]=port(b,0,kill) ;
		pair dir=rotate(90)*dir(b,.5) ;
		c.port[finner(i)]=port(shift(.1dir)*b,1,
			forwards=false,dir=(0,1),length=defaultportlength-.1,kill) ;
	}

	return c ;
}


void test_box1(picture pic=currentpicture) {
	render(pic,box(.5,"$!$",.2,rectangle(width=3,height=2,round=.35))) ;
}

addtest("test_box1: basic box generator.",test_box1,"box") ;


cell box (explicit pair A=(0,0), real main, real w=2, real h=2,
		real margin=defaultboxmargin, pair dir=S,
		Label L="!", real radius=defaultboxradius) {
	rectangle r = rotate(degrees(dir)+90,A)*
		shift(((normalize(main))*(w-2margin)+margin)*W)*
		rectangle(A,width=w,height=h,angle=0,round=0,
		margin=margin) ;
	return box(main,L,radius,r) ;
}


void test_box2(picture pic=currentpicture) {
	dot(pic,E+S) ;
	render(pic,box(A=S+E,.2,w=3,dir=S+E)) ;
}

addtest("test_box2: box by dimensions and direction.",test_box2,"box") ;

pair boxdir(cell c,int i) {
	return rotate(90)*unit(point(c@fouter(i),1)-point(c@fouter(i),0)) ;
}

path boxbase(cell c, int i) {
	pair A,B ; 
	A=point(c@fouter(i),0) ; 
	B=proj(A,point(c@finner(i),0),boxdir(c,i)) ; 
	return A--B ;
}

pair boxpoint(cell c, int i, real x) {
	return(point(boxbase(c,i),x)) ; 
}

real boxabs(cell c, int i, pair C) {
	pair A=boxpoint(c,i,0) ;
	pair B=boxpoint(c,i,1) ;
	if(A==B) 
		return 0 ; 
	return(dot(C-A,B-A)/dot(B-A,B-A)) ;
}

pair boxproj(cell c, int i, pair A) {
	return point(boxbase(c,i),boxabs(c,i,A)) ;
}

link outer(cell c, int i, real x, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=true, type ts[]) {
	link l=port(boxpoint(c,i,x),rotate(-90)*boxdir(c,i),length,arrow,forwards) ;
	l.type=ts;
	return l;
}
link inner(cell c, int i, real x, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=false, type ts[]) {
	pair dir=rotate(90)*boxdir(c,i);
	link l=port(boxpoint(c,i,x)+.1dir,dir,length-.1,arrow,forwards) ;
	l.type=ts;
	return l;
}
void boxport(cell c, int i, real x, 
		real lengthout=defaultportlength, real lengthin=defaultportlength,
		bool arrowout=linkarrow, bool arrowin=linkarrow, 
		bool forwardsout=true, bool forwardsin=false,
		type tsin[], type tsout[]) {
			c.port.push(outer(c,i,x,lengthout,arrowout,forwardsout,tsout)) ;
			c.port.push(inner(c,i,x,lengthin,arrowin,forwardsin,tsin)) ;
}

link outer(cell c, int i, real x, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=true ... Label []Ls) {
	return outer(c,i,x,length,arrow,forwards,types(Ls,arrow,forwards)) ;
}

link inner(cell c, int i, real x, real length=defaultportlength,
	bool arrow=linkarrow, bool forwards=false ... Label []Ls) {
	return inner(c,i,x,length,arrow,forwards,types(Ls,arrow,forwards)) ;
}
void boxport(cell c, int i, real x, 
		real lengthout=defaultportlength, real lengthin=defaultportlength,
		bool arrowout=linkarrow, bool arrowin=linkarrow, 
		bool forwardsout=true, bool forwardsin=false
		... Label[]Ls) {
			c.port.push(outer(c,i,x,lengthout,arrowout,forwardsout ... Ls)) ;
			c.port.push(inner(c,i,x,lengthin,arrowin,forwardsin ... Ls)) ;
}


link projout(cell c, int i, pair A, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=true, type ts[]) {
	link l=port(boxproj(c,i,A),rotate(-90)*boxdir(c,i),length,arrow,forwards) ;
	l.type=ts;
	return l;
}
link projin(cell c, int i, pair A, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=false, type ts[]) {
	pair dir=rotate(90)*boxdir(c,i);
	link l=port(boxproj(c,i,A)+.1dir,dir,length-.1,arrow,forwards) ;
	l.type=ts;
	return l;
}
void boxport(cell c, int i, explicit pair A, 
		real lengthout=defaultportlength, real lengthin=defaultportlength,
		bool arrowout=linkarrow, bool arrowin=linkarrow, 
		bool forwardsout=true, bool forwardsin=false,
		type tsin[], type tsout[]) {
			c.port.push(projout(c,i,A,lengthout,arrowout,forwardsout,tsout)) ;
			c.port.push(projin(c,i,A,lengthin,arrowin,forwardsin,tsin)) ;
}

link projout(cell c, int i, pair A, real length=defaultportlength,
	bool arrow=linkarrow, bool forwards=true ... Label []Ls) {
	return projout(c,i,A,length,arrow,forwards,types(Ls,arrow,forwards)) ;
}
link projin(cell c, int i, pair A, real length=defaultportlength, 
	bool arrow=linkarrow, bool forwards=false ... Label []Ls) {
	return projin(c,i,A,length,arrow,forwards,types(Ls,arrow,forwards)) ;
}
void boxport(cell c, int i, explicit pair A,
		real lengthout=defaultportlength, real lengthin=defaultportlength,
		bool arrowout=linkarrow, bool arrowin=linkarrow, 
		bool forwardsout=true, bool forwardsin=false
		... Label[]Ls) {
			c.port.push(projout(c,i,A,lengthout,arrowout,forwardsout ... Ls)) ;
			c.port.push(projin(c,i,A,lengthin,arrowin,forwardsin ... Ls)) ;
}
void boxaux(cell c, int i,
		real lengthout=defaultportlength, real lengthin=defaultportlength
		... link[]ls) {
		for (int j=0;j<ls.length;++j) {
			link l=ls[j] ; 
			boxport(c,i,endpoint(l),lengthout,lengthin) ; 
			int n=boxnumports(c)-1 ; 
			cut(l,c@inner(n)) ; 
			link lo=c@outer(n) ; 
			lo.forwards=l.forwards ; 
			lo.arrow=l.arrow ; 
			lo.type=map(new type (type t) { return t.copy() ; }, (type[])l.type) ;
		}
}


void test_boxproj(picture pic=currentpicture) {
	access tricell ;
	dot(pic,(0,0)) ;
	cell c = tricell.tricell("c","","","") ;
	cell b = box(S,.75,w=3,h=2) ;
	cut(c@0,b@inner(0)) ;
	cut(c@1,projin(b,0,W)) ;
	cut(c@2,projin(b,2,endpoint(c@2))) ;
	render(pic,b,c) ; 
}
addtest("test_boxproj: box ports as projections on sides.",
	test_boxproj,"box","boxproj") ;

void boxports(cell c, int i, int num, real start=0, real end=1, 
	bool rev=false,
	real lengthout=defaultportlength, real lengthin=defaultportlength,
	bool arrowout=linkarrow, bool arrowin=linkarrow, 
	bool forwardsout=true, bool forwardsin=false) {
			if (num==1) 
				boxport(c,i,start,lengthout,lengthin,
					arrowout,arrowin,forwardsout, forwardsin) ;
			else if (num>1) {
				real s=rev?end:start ;
				real e=rev?start:end ;
				for (int j=0;j<num;++j) 
					boxport(c,i,s+j*(e-s)/(num-1),lengthout,lengthin,
						arrowout,arrowin,forwardsout,forwardsin) ;
			}
}


void test_boxports(picture pic=currentpicture) {
	access tricell ;
	dot(pic,(0,0)) ;
	cell c = tricell.tricell("c","","","") ;
	cell b = box(S,.75,w=3,h=2) ;
	boxports(b,0,3,end=.5) ;
	cut(c@0,b@inner(0)) ;
	cut(c@1,b@inner(3)) ;
	cut(c@2,b@inner(1)) ;
	render(pic,b,c) ; 
}
addtest("test_boxports: series of box ports.",
	test_boxports,"box","boxports") ;


