import common ;
import tests ;
import link ;
import render ;

/**************************************
 *                                    * 
 *             Cell                   * 
 *                                    * 
 **************************************/

/* Basic cell structure */

struct cell {
	picture pic ;
	link port[] ; 
	void transform (transform T) {
		pic = T * pic ; 
		for (int i=0; i<port.length; ++i)
			port[i].transform(T) ;
	}
	void render(picture pic=currentpicture) {
		for (int i=0; i<port.length; ++i) 
			port[i].render(pic) ;
		add(pic, this.pic) ;
	}
	void init (picture pic, link port[]) { 
		this.pic = pic ;
		this.port = port ;
	}
	cell copy () { 
		cell c = new cell ;
		link p[] = map(copy,port) ;
		c.init(pic.copy(), p) ; 
		return c ;
	}
}

cell operator init () { return new cell ; }

cell cell(picture pic, link port[]=new link[]) {
	cell c = new cell ;
	c.init(pic,port) ;
	return c ;
}

render operator cast (cell c) {
	render r = new render ;
	r.render = c.render ;
	return r ;
}

/* Cell operations */

link operator @ (cell c, int i) { return c.port[i] ; } 

cell copy(cell c) { return c.copy() ; } 

cell operator * (transform T, cell c) { 
	cell cc = copy(c) ; cc.transform(T) ; return cc ; 
}

void test_cell (picture pic=currentpicture) {
	unitsize(pic,2cm) ;
	cell c ;
	draw(c.pic,unitsquare) ;
	c.port.push(link((0,0)--S)) ;
	c.port.push(link(E--E+S)) ;
	render(pic,c) ;
	c = rotate(45)*shift(2E)*c ;
	render(pic,c) ;
}

addtest("test_cell: basic cell.",test_cell,"cell") ;

/* Link operations on cell ports */

link merge(cell c, int p=0, cell d, int q=0, 
		linkmerge merge = defaultlinkmerge) {
	link p = c.port[p] ;
	link q = d.port[q] ;
	return(merge(p,reverse(q),merge));
}

void cut (cell c, int p=0, cell d, int q=0, 
		linkmerge m1 = defaultlinkmerge, linkmerge m2=Discard) {
	cut (c.port[p],d.port[q],m1,m2) ;
}

/* Functions to create ports */ 

link port (pair A, real angle, real length=defaultportlength, 
		bool arrow=linkarrow, bool forwards=true ... Label [] L) {
	return link(A--A+length*expi(angle),arrow,forwards,L);
}

link port (pair A, pair dir, real length=defaultportlength, 
		bool arrow=linkarrow, bool forwards=true ... Label [] L) {
	return port(A,angle(dir),length,arrow,forwards ... L) ;
}

link port (path g, position pos, pair dir=S, real length=defaultportlength, 
		bool arrow=linkarrow, bool forwards=true ... Label [] L) {
	return port(point(g,pos),rotate(degrees(g,pos))*dir,length,
		arrow,forwards ... L) ;
}

link [] ports (path g, int num, pair dir=S, position begin=BeginPoint, 
		position end=EndPoint, real length=defaultportlength, 
		bool arrow=linkarrow, bool forwards=true) {
	path g = subpath(g,begin,end) ;
	link l[] ;
	if (num == 1)
		l.push(port(
			g,
			MidPoint,
			length=length,
			dir=dir,
			arrow=arrow,
			forwards=forwards)) ; 
	else if (num > 1)
		for (int i = 1 ; i <= num ; ++i)
			l.push(port(
				g,
				Relative((i-1)/(num-1)),
				length=length,
				dir=dir,
				arrow=arrow,
				forwards=forwards)) ;
	return l ; 
}


void test_port (picture pic=currentpicture) { 
	unitsize(2cm) ; 
	link l3 = port(2 E+2S,S+E,"l_3") ;
	link l4 = port(2 E+2S,S+W,"l_4") ;
	render(pic,l3,l4) ;
	link l5 = port(2E+S,E,"A") ; 
	link l6 = port(3E+2S,E,"B") ; 
	link l7=merge(l5,l6) ;
	render(pic,l5,l6,l7) ;
	render(pic,l7) ;
}

addtest("test_port: ports from base point; merging.",
	test_port,"port","merge") ;

void test_ports (picture pic=currentpicture) { 
	unitsize(2cm) ; 
	path g = W--E{N}..N ;
	link l[] = ports(g,4) ; 
	render(pic,g,l[0],l[1],l[2],l[3]) ;
}

addtest("test_ports: ports from base path.",test_ports,"port","ports") ;


