import diffnets ; 

Label emptystring (int) { return "@none@" ; }

cell[] commzone_array (int n=3) {

	real w = defaultcellwidth+.05n ;
	real r = n*(w+.1)/pi ; 
	triangle t = triangle(base=w) ;
	cell c = tricell(cont,t,reverse=true,Label("") ... sequence(emptystring,n-1));
	cell m = tricell(mult,t,reverse=false,Label("") ... sequence(emptystring,n-1));
	cell con [] ; 
	cell coc [] ; 
	
	con.cyclic(true) ;
	coc.cyclic(true) ;
	
	for (int i=0; i<n ; ++i) {
		real a = i/n*360;
		pair d = rotate(a)*S ; 
		pair o = rotate(-90)*d ; 
		con.push(shift(r*d+(w-.5defaultround)/2*o)*(rotate(a)*c)) ; 
		coc.push(shift(r*d-(w-.5defaultround)/2*o)*(rotate(a)*m)) ; 
	}

	for (int i=0; i<n ; ++i) {
		for (int j=1; j<n ; ++j) {
			cut(con[i].port[j],coc[i-j].port[n-j]);
		}
	}

	con.append(coc) ;
	return con ;
}

void test_commzone_array (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	int n = 7 ; 
	cell comm [] = commzone_array(n) ; 
	for (int i=0;i<2n;++i) {
		render(pic,comm[i]) ;
	}
}

addtest("test_commzone_array: communication zone (`cell []' version).",test_commzone_array,"commzone") ;

cell commzone (int n=3) {
	cell ca[] = commzone_array(n) ; 
	cell c = new cell ; 
	for (int i=0;i<n;++i) {
		add(c.pic,ca[i].pic) ; 
		add(c.pic,ca[i+n].pic) ; 
		for (int j=1;j<n;++j) {
			render(c.pic,ca[i].port[j]) ;
			render(c.pic,ca[i+n].port[j]) ;
		}
		c.port.push(ca[i].port[0]) ;
		c.port.push(ca[i+n].port[0]) ;
	}
	return c ;
}

void test_commzone (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	render(commzone(9)) ; 
}

addtest("test_commzone: communication zone (`cell' version).",test_commzone,"commzone") ;

void test_commzones (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	cell c9 = rotate(90)*commzone(9) ; 
	cell c7 = shift(endpoint(c9.port[0])+endpoint(c9.port[1]))*rotate(-90)*commzone(7) ; 
	cut(c9,0,c7,1) ;
	cut(c9,1,c7,0) ;
	render(c9,c7) ; 
}

addtest("test_commzone: communication zone (`cell' version).",test_commzone,"commzone") ;

cell comm (int n=3,real w=defaultcellwidth) {
	
	cell c = new cell ; 

	if (n<=0) return c ; 

	if (n==1) {
		pair A = w/2*W ;
		pair B = w/2*E ;
		draw(c.pic,A--B{N}..{S}A) ; 
		c.port.push(port(interp(A,B,defaultround/w),S,forwards=false,"")) ; 
		c.port.push(port(interp(B,A,defaultround/w),S,forwards=true,"")) ; 
		return c ;
	}

	real r = n*w/pi ; 
	
	pair D [] ; 
	pair A [] ; 
	pair B [] ; 

	A.cyclic(true) ; 
	B.cyclic(true) ; 
	D.cyclic(true) ; 
	
	for (int i=0; i<n ; ++i) {
		real a = i/n*360;
		pair d = rotate(a)*S ; 
		pair o = rotate(-90)*d ; 
		A.push(r*d+w/2*o) ; 
		B.push(r*d-w/2*o) ; 
		D.push(d) ; 
	}

	for (int i=0; i<n ; ++i) {
		draw(c.pic,A[i]--B[i]) ; 
		pair z = A[i+1]-B[i] ; 
		draw(c.pic,B[i]{rotate(30)*z}..{rotate(-30)*z}A[i+1]) ; 
		c.port.push(port(interp(A[i],B[i],defaultround/w),D[i],forwards=false,"")) ; 
		c.port.push(port(interp(B[i],A[i],defaultround/w),D[i],forwards=true,"")) ; 
	}

	return c ;
}

void test_comm (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	int n = 7 ; 
	cell comm = comm(n) ; 
	render(pic,comm) ;
}

addtest("test_comm: communication zone.",test_comm,"comm") ;




