import type ;
import tricell ;
import rectcell ; 
import boxes ; 
import reductions ; 

usepackage("cmll") ;

unitsize(.8cm) ;
defaultpen(fontsize(7pt)) ; 

/**************************************
 *                                    * 
 *    Linear logic cells              * 
 *                                    * 
 **************************************/


Label tensor = Label(embed=Shift,"\otimes") ;
Label par    = Label(embed=Shift,"\parr") ;
Label one    = Label(embed=Shift,"\mathbf 1") ;
Label bot    = Label(embed=Shift,"\bot") ;
Label mult   = Label(embed=Shift,"\mathrm m") ;
Label unit   = Label(embed=Shift,"\mathrm u") ;
Label cont   = Label(embed=Shift,"\mathrm c") ;
Label weak   = Label(embed=Shift,"\mathrm w") ;
Label der    = Label(embed=Shift,"\mathrm d") ;
Label cod    = Label(embed=Shift,"\partial") ;

Label postree   = Label(embed=Shift,"+") ;
Label negtree   = Label(embed=Shift,"-") ;
Label strtree   = Label(embed=Shift,"\mathrm c") ;
Label costree   = Label(embed=Shift,"\mathrm m") ;

void kill(link l) {
	l.type(kill) ;
}

void ellip(link l) {
	l.type(ellip) ;
}

void untype(link l) {
	l.type=new type[] ;
}

void unrender(link l) {
	untype(l) ; 
	kill(l) ;
}

cell quick_cell(Label L, bool reverse=false ... string [] s  ) {
	Label Ls [] = new Label [] ;
	for (int i=0 ; i<s.length ; ++i) 
		Ls.push(Label(s[i], embed=Shift)) ;
	return tricell(L,reverse ... Ls) ;
}


cell tensor(string A="", string B="", string C="") {
	if (A+B!="" && C=="")
		C=A+" \otimes "+B ;
	return(quick_cell(tensor,C,A,B)) ;
}

cell one(string A="\mathbf 1") {
	return(quick_cell(one,A)) ;
}

cell par(string A="", string B="", string C="") {
	if (A+B!="" && C=="")
		C=A+" \parr "+B ;
	return(quick_cell(par,C,A,B)) ;
}

cell bot(string A="\bot") {
	return(quick_cell(bot,A)) ;
}

cell mult(string A="", string B="", string C="") {
	if (B+C=="")
		B=C=A ; 
	return(quick_cell(mult,A,B,C)) ;
}

cell unit(string A="") {
	return(quick_cell(unit,A)) ;
}

cell cont(string A="", string B="", string C="") {
	if (B+C=="")
		B=C=A ; 
	return(quick_cell(cont,A,B,C)) ;
}

cell weak(string A="") {
	return(quick_cell(weak,A)) ;
}

cell der(string A="", string B="") {
	if (A!="" && B=="")
		B="\wn "+A ;
	return(quick_cell(der,B,A)) ;
}

cell cod(string A="", string B="") {
	if (A!="" && B=="")
		B="\oc "+A ;
	return(quick_cell(cod,B,A)) ;
}

cell rtensor(string A="", string B="", string C="") {
	if (A+B!="" && C=="")
		C=B+" \parr "+A ;
	return(quick_cell(tensor,true,C,A,B)) ;
}

cell rone(string A="\bot") {
	return(quick_cell(one,true,A)) ;
}

cell rpar(string A="", string B="", string C="") {
	if (A+B!="" && C=="")
		C=A+" \otimes "+B ;
	return(quick_cell(par,true,C,B,A)) ;
}

cell rbot(string A="\mathbf 1") {
	return(quick_cell(bot,true,A)) ;
}

cell rmult(string A="", string B="", string C="") {
	if (B+C=="")
		B=C=A ; 
	return(quick_cell(mult,true,A,B,C)) ;
}

cell runit(string A="") {
	return(quick_cell(unit,true,A)) ;
}

cell rcont(string A="", string B="", string C="") {
	if (B+C=="")
		B=C=A ; 
	return(quick_cell(cont,true,A,B,C)) ;
}

cell rweak(string A="") {
	return(quick_cell(weak,true,A)) ;
}

cell rder(string A="", string B="") {
	if (A!="" && B=="")
		B="\oc "+A ;
	return(quick_cell(der,true,B,A)) ;
}

cell rcod(string A="", string B="") {
	if (A!="" && B=="")
		B="\wn "+A ;
	return(quick_cell(cod,true,B,A)) ;
}

void test_diffnet_cells (picture pic = currentpicture) {
	unitsize(pic,1cm) ;
	cell t=tensor("A","B") ;
	cell p=shift(2,0)*par("A^\bot","B^\bot") ;
	cell pp=shift(6,0)*par("A^\bot","B^\bot") ;
	cut(t,p) ;
	render(pic,t,p,pp) ;
}

addtest("test_diffnet_cells: example cells with typing.",
	test_diffnet_cells,"file:diffnets.asy","tricell","cut") ;

/**************************************
 *                                    * 
 *           Utility                  * 
 *                                    * 
 **************************************/

/* Naming ports */

void varname (Label L, path g) {
	label(Label(ensuremath(L),align=Relative(N),position=Relative(1)),g) ;
}

void varname (Label L, link l) {
	type(l,type(Label(L,position=Relative(1),align=Relative(N)),arrow=false)) ;
}

void varname (link l, Label L) {
	varname(L,l) ;
}

/**************************************
 *                                    * 
 *           Generic net              * 
 *                                    * 
 **************************************/

cell lblob (pair A=(0,0), Label L="", string Lo, 
		real w=2, real h=1, pair dir=S ... type types[]) {
	return rectcell(A+rectround*(rotate(-90)*unit(dir)),
				w,h,L,type(Lo), after=types, dir=dir) ;
}
cell lblob (pair A=(0,0), Label L="", string Lo, 
		real w=2, real h=1, pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return lblob(A,L,Lo,w,h,dir ... types) ;
}

cell rblob (pair A=(0,0), Label L="", string Lo, 
		real w=2, real h=1, pair dir=S ... type types[]) {
	cell c=rectcell(A+(-rectround+w)*(rotate(-90)*unit(dir)),
				w,h,L,type(Lo),before=types, dir=dir) ;
	c.port[0].type[0].L.align=LeftSide ;
	return c ;
}
cell rblob (pair A=(0,0), Label L="", string Lo, real w=2, real h=1,
		pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return rblob(A,L,Lo,w,h,dir ... types) ;
}

cell dblob (pair A=(0,0), Label L="", string Lo, real w=2, real h=1,
		pair dir=S ... type types[]) {
	return rectcell(A+(w/2)*(rotate(-90)*unit(dir)),
				w,h,L,type(Lo),top=types,dir=dir) ;
}

cell dblob (pair A=(0,0), Label L="", string Lo, real w=2, real h=1,
		pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s,forwards=false) ;
	return dblob(A,L,Lo,w,h,dir ... types) ;
}

cell bblob (pair A=(0,0), Label L="", Label Lo, int numaux,
		real w=2, real h=1, pair dir=S ... type types[]) {
	type aux[] = types[sequence(numaux)];
	type var[] = types[sequence(numaux,types.length-1)] ;
	return rectcell(A+rectround*(rotate(-90)*unit(dir)),
				w,h,L,type(Lo),after=aux,top=var,dir=dir) ;
}

cell bblob (pair A=(0,0), Label L="", Label Lo, int numaux, 
		real w=2, real h=1, pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s[sequence(numaux)]) ;
	types.append(types(s[sequence(numaux,s.length-1)],forwards=false)) ;
	return bblob(A,L,Lo,numaux,w,h,dir ... types) ;
}

void test_blobs (picture pic=currentpicture) {
	unitsize(pic,1cm) ; 
	render(pic,lblob(w=3,"s","A\parr B","\wn A^\bot",ellip,"\oc B"));
	render(pic,rblob(A=3W,w=3,dir=W,"s","A\parr B","\wn A^\bot",ellip,"\oc B"));
	render(pic,dblob(A=2S,dir=S,w=3,"s","A\parr B","\wn A^\bot",ellip,"\oc B"));
	dot(pic,0) ;
	dot(pic,3W) ;
	dot(pic,2S) ;
}

addtest("test_blobs: blobs with transforms.",
	test_blobs,"file:diffnets.asy","blob","lblob","rblob","dblob") ;

/**************************************
 *                                    * 
 *              Boxes                 * 
 *                                    * 
 **************************************/


cell box (explicit pair A=(0,0), real w=2, real h=2, real x=0,
		type mainout,
		type mainin=type(Label(""),forwards=false), 
		type before[] = new type[], 
		type after[] = new type[],
		type right[] = new type[],
		type top[] = new type[],
		type left[] = new type[],
		real margin=defaultboxmargin, pair dir=S, 
		Label L="!", real radius=defaultboxradius) {
	int Before=before.length;
	int After=after.length;
	int Right=right.length;
	int Top=top.length;
	int Left=left.length;

	cell c=box(A,x,w,h,margin,dir,L,radius) ;

	if (After>0) 
		boxports(c,0,After,1,x+(1/After)*(1-x),rev=true) ;
	if (Right==1) 
		boxports(c,1,1,.5) ;
	else 
		boxports(c,1,Right) ;
	if (Top==1) 
		boxports(c,2,1,.5) ;
	else 
		boxports(c,2,Top) ;
	if (Left==1) 
		boxports(c,3,1,.5) ;
	else 
		boxports(c,3,Left) ;
	if (Before>0) 
		boxports(c,0,Before,0,x*(1-1/Before)) ;

	c.port[outer(0)].type(mainout) ;
	c.port[inner(0)].type(mainin) ;

	type out[]=new type[] {} ;
	out.append(after) ;
	out.append(right) ;
	out.append(top) ;
	out.append(left) ;
	out.append(before) ;
	for (int i=0; i<out.length; ++i) 
		c.port[outer(i+1)].type(out[i]) ;

	for (int i=0; i<out.length; ++i) {
		link p=c.port[inner(i+1)];
		p.type(reverse(out[i],p.g)) ;
	}
	return c ;
}

void test_box3(picture pic=currentpicture) {
	render(pic,box(w=3,h=1.1,x=.25,dir=S+E,type("!A"),type("A"),
			top=types("",ellip,""),after=types("","",""))) ;
}

addtest("test_box3: box with types.",test_box3,"box","type","types") ;

real defaultraise = defaultportlength+.1 ;

cell lbox (pair A=(0,0), string typeout, real w=2, real h=1, 
		real margin=defaultboxmargin, real raise=defaultraise,
		pair dir=S ... type types[]) {
	
	real width = w+2*margin ;
	real height = h+margin+raise ;

	return box(A+raise*unit(dir),
		width,height,0,
		type(Label("\oc "+typeout,align=RightSide)),
		type(Label(typeout,align=LeftSide),forwards=false),
		after=types,margin=margin,dir=dir) ;
}

cell lbox (pair A=(0,0), string typeout, real w=2, real h=1,
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S, string L ... string Ls[]) {

	string s[] ;
	s.push(L) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return lbox(A,typeout,w,h,margin,raise,dir ... types) ;
}

void test_lbox (picture pic=currentpicture) { 
	unitsize(pic, 1cm) ; 
	render(pic, lbox(A=5W,w=3,"A\parr B","\wn A^\bot",ellip,"\oc B"));
	render(pic, lbox(A=2S,dir=S+E,w=3,"A\parr B","\wn A^\bot",ellip,"\oc B"));
	dot(pic,2S) ;
}

addtest("test_lbox",test_lbox,"file:diffnets.asy","box","lbox") ;

cell rbox (pair A=(0,0), string typeout, real w=2, real h=1, 
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S ... type types[]) { 
	
	real width = w+2*margin ;
	real height = h+margin+raise ;

	return box(A+raise*unit(dir), width,height,x=1,
		type(Label("\oc "+typeout,align=LeftSide)),
		type(typeout,forwards=false),
		before=types,margin=margin,dir=dir) ;
}

cell rbox (pair A=(0,0), string typeout, real w=2, real h=1,
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S, string L ... string Ls[]) {
	string s[] ;
	s.push(L) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return rbox(A,typeout,w,h,margin,raise,dir ... types) ;
}

void test_rbox (picture pic=currentpicture) { 
	unitsize(pic, 1cm) ; 
	render(pic, rbox(A=2S,dir=S+E,w=3,"A\parr B","\wn A^\bot",ellip,"\oc B"));
	dot(pic,2S) ;
}

addtest("test_rbox",test_rbox,"file:diffnets.asy","box","rbox") ;

cell dbox (pair A=(0,0), string typeout, real w=2, real h=1, 
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S ... type types[]) {  
	
	real width = w+2*margin ;
	real height = h+margin+raise ;

	return box(A+raise*unit(dir),
		width,height,.5,
		type(Label("\oc "+typeout)),
		type(Label(typeout,align=LeftSide),forwards=false),
		top=types,margin=margin,dir=dir) ;

}

cell dbox (pair A=(0,0), string typeout, real w=2, real h=1,
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S, string L ... string Ls[]) {
	string s[] ;
	s.push(L) ;
	s.append(Ls) ;
	type types []=types(s,forwards=false) ;
	return dbox(A,typeout,w,h,margin,raise,dir ... types) ;
}

cell bbox (pair A=(0,0), string typeout, int numaux, real w=2, real h=1, 
		real margin=defaultboxmargin, real raise=defaultraise, 
		pair dir=S ... type types[]) {  
	
	type aux[] = types[sequence(numaux)];
	type var[] = types[sequence(numaux,types.length-1)] ;

	real width = w+2*margin ;
	real height = h+margin+raise ;

	return box(A+raise*unit(dir),
		width,height,(numaux==0)?.5:0,
		type(Label("\oc "+typeout)),
		type(Label(typeout,align=LeftSide),forwards=false),
		top=var,after=aux,margin=margin,dir=dir) ;

}

cell bbox (pair A=(0,0), string typeout, int numaux, real w=2, real h=1,
		real margin=defaultboxmargin, real raise=defaultraise,
		pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s[sequence(numaux)]) ;
	types.append(types(s[sequence(numaux,s.length-1)],forwards=false)) ;
	return bbox(A,typeout,numaux,w,h,margin,raise,dir ... types) ;
}

void test_dbox (picture pic=currentpicture) { 
	unitsize(pic, 1cm) ; 
	render(pic, dbox(A=2S,dir=S+E,w=3,"A\parr B","\wn A^\bot",ellip,"\oc B"));
	dot(pic,2S) ;
}

addtest("test_dbox",test_dbox,"file:diffnets.asy","box","dbox") ;

void test_boxes(picture pic=currentpicture) {
	unitsize(pic,1.5cm) ;
	cell b =lbox(w=3,"P","N_1","N_2","N_3") ;
	cut(b@inner(0),b@inner(1)) ;
	cell c = shift(2E)*tricell(1,"d") ;
	cell l = shift(3E)*vert() ;
	cut(c,0,b,inner(2)) ;
	cut(c,1,l,1) ;
	cut(l,0,b,inner(3)) ;
	cell r =rbox(8E,dir=S+W,w=3,"P","N_1","N_2","N_3") ;
	dot(pic,8E) ;
	render(pic,b,l,c,r) ;
}

addtest("test_boxes",test_boxes,"file:diffnets.asy","box","lbox","rbox") ;
cell lboxed (pair A=(0,0), Label L="", string typeout,
		real w=2, real h=1, real raise=defaultraise, pair dir=S ... type types[]) {

	real margin=defaultboxmargin ;
	picture pic ; 

	cell b = lbox(A,typeout,w-2*rectround,h,
			margin=margin+rectround,raise=raise,dir=dir ... types) ;
	cell s = lblob(A,L,typeout,w,h,dir=dir ... types) ;
	for (int i=0; i<=types.length;++i) {
		cut(s,i,b,inner(i)) ;
		render(pic,b.port[inner(i)]) ;
	}
	render(pic, s) ;
	render(pic, b.pic) ;
	link ports[]=new link[] ;
	for (int i=0; i<=types.length;++i) 
		ports.push(b.port[outer(i)]) ;
	return cell(pic,ports) ;
}
cell lboxed (pair A=(0,0), Label L="", string typeout, 
	real w=2, real h=1, real raise=defaultraise,
	pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return lboxed(A,L,typeout,w,h,raise,dir ... types) ;
}

	
void test_lboxed (picture pic=currentpicture) { 
	unitsize(pic,1cm) ; 
	render(pic,lboxed("s","A",2,1,dir=S+E,"",ellip,"") ) ;
	render(pic,lboxed(4E,"s","A",2,1,dir=W,"",ellip,"") ) ;
}

addtest("test_lboxed",test_lboxed,"file:diffnets.asy",
	"blob","lblob","box","lbox","boxed","lboxed") ;

cell rboxed (pair A=(0,0), Label L="", string typeout,
		real w=2, real h=1, real raise=defaultraise, pair dir=S ... type types[]) {

	real margin=defaultboxmargin ;
	picture pic ; 

	cell b = rbox(A,typeout,w-2*rectround,h,margin=margin+rectround,raise=raise,
			dir=dir ... types) ;
	cell s = rblob(A,L,typeout,w,h,dir=dir ... types) ;
	for (int i=0; i<=types.length;++i) {
		cut(s,i,b,inner(i)) ;
		render(pic,b.port[inner(i)]) ;
	}
	render(pic, s) ;
	render(pic, b.pic) ;
	link ports[]=new link[] ;
	for (int i=0; i<=types.length;++i) 
		ports.push(b.port[outer(i)]) ;
	return cell(pic,ports) ;
}
cell rboxed (pair A=(0,0), Label L="", string typeout, 	real w=2, real h=1, 
		real raise=defaultraise, pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s) ;
	return rboxed(A,L,typeout,w,h,raise,dir ... types) ;
}

void test_rboxed (picture pic=currentpicture) { 
	unitsize(pic,1cm) ; 
	render(pic,rboxed("s","A",2,1,dir=S+E,"",ellip,"") ) ;
	render(pic,rboxed(4E,"s","A",2,1,dir=W,"",ellip,"") ) ;
}

addtest("test_rboxed",test_rboxed,"file:diffnets.asy",
		"blob","rblob","box","rbox","boxed","rboxed") ;
	
cell dboxed (pair A=(0,0), Label L="", string typeout,
		real w=2, real h=1, real raise=defaultraise, pair dir=S ... type types[]) {

	real margin=defaultboxmargin ;
	picture pic ; 

	cell b = dbox(A,typeout,w-2*rectround,h,margin=margin+rectround,raise=raise,
			dir=dir ... types) ;
	cell s = dblob(A,L,typeout,w,h,dir=dir ... types) ;
	for (int i=0; i<=types.length;++i) {
		cut(s,i,b,inner(i)) ;
		render(pic,b.port[inner(i)]) ;
	}
	render(pic, s) ;
	render(pic, b.pic) ;
	link ports[]=new link[] ;
	for (int i=0; i<=types.length;++i) 
		ports.push(b.port[outer(i)]) ;
	return cell(pic,ports) ;
}
cell dboxed (pair A=(0,0), Label L="", string typeout, real w=2, real h=1, 
		real raise=defaultraise, pair dir=S, string LL ... string Ls[]) {

	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s,forwards=false) ;
	return dboxed(A,L,typeout,w,h,raise,dir ... types) ;
}

	
void test_dboxed (picture pic=currentpicture) { 
	unitsize(pic,1cm) ; 
	render(pic,dboxed("s","A",2,1,dir=S+E,"",ellip,"") ) ;
	render(pic,dboxed(4E,"s","A",2,1,dir=W,"",ellip,"") ) ;
}

cell bboxed (pair A=(0,0), Label L="", string typeout, int numaux, real w=2, 
		real h=1, real raise=defaultraise, pair dir=S ... type types[]) {

	real margin=defaultboxmargin ;
	picture pic ; 

	cell b = bbox(A,typeout,numaux=numaux,w-2*rectround,h,
			margin=margin+rectround,raise=raise,dir=dir ... types) ;
	pair sh = numaux==0 ? rotate(-90)*(w-2rectround)/2*dir: 0 ;
	cell s = shift(sh)*bblob(A,L,typeout,numaux,w,h,dir=dir ... types) ;
	for (int i=0; i<=types.length;++i) {
		cut(s,i,b,inner(i)) ;
		render(pic,b.port[inner(i)]) ;
	}
	render(pic, s) ;
	render(pic, b.pic) ;
	link ports[]=new link[] ;
	for (int i=0; i<=types.length;++i) 
		ports.push(b.port[outer(i)]) ;
	return cell(pic,ports) ;
}

cell bboxed (pair A=(0,0), Label L="", string typeout, int numaux, real w=2, 
		real h=1, real raise=defaultraise, pair dir=S, string LL ... string Ls[]) {
	string s[] ;
	s.push(LL) ;
	s.append(Ls) ;
	type types []=types(s[sequence(numaux)]) ;
	types.append(types(s[sequence(numaux,s.length-1)],forwards=false)) ;
	return bboxed(A,L,typeout,numaux,w,h,raise,dir ... types) ;
}

addtest("test_dboxed",test_dboxed,"file:diffnets.asy",
	"blob","dblob","box","dbox","boxed","dboxed") ;

