import common ;
import tests ;
import render ; 
import type ;

/**************************************
 *                                    * 
 *             Link                   * 
 *                                    * 
 **************************************/

/* Basic link structure */

struct link {

	path g ; 
	bool arrow ;
	bool forwards ;
	type type[] ;

	void init(path g, bool arrow=linkarrow, bool forwards=true,
			type type[]=new type[]) {
		this.g=g ;
		this.arrow=arrow ;
		this.forwards=forwards ;
		this.type=type ;
	}
	init(g) ;

	void init(link l) { 
		init(l.g,l.arrow,l.forwards,l.type=map(copy,l.type)) ; 
	}
	
	link copy() {
		link l = new link ;
		l.init(g,arrow,forwards,map(copy,type)) ;
		return l ;
	}

	void type(type t) {
		type.push(t) ;
	}
	
	// workaround name clash with function 'type' in type.asy
	type Labeltype(Label L, bool arrow, bool forwars) {
		return type(L,arrow,forwards) ;
	}

	void type(Label L,bool arrow=arrow, bool forwards=forwards) {
		type(Labeltype(L,arrow,forwards)) ;
	}

	void type(Label L[],bool arrow=arrow, bool forwards=forwards) {
		for (int i=0 ; i < L.length ; ++i) 
			type(L[i],arrow,forwards) ;
	}

	void transform(transform T) {
		g = T*g ;
	}

	void render(picture pic=currentpicture) {
		bool dorender = true ;
		for (int i=0; i<type.length; ++i) {
			type[i].out(pic,g) ;
			if (hastag(type[i].L,kill)) { dorender=false ; } 
		}
		if (dorender) 
			draw(pic,g) ;
 	} 

}

link operator init () {
	return new link ;
}

path operator cast (link l) { return l.g ; }

path path(link l) { return l.g ; }

render operator cast (link l) {
	render r = new render ;
	r.render = l.render ;
	return r ;
}

link link(path g, bool arrow=linkarrow, bool forwards=true, type type[]) {
	link l ; 
	l.init(g,arrow,forwards,type) ;
	return l ;
}

link link(path g, bool arrow=linkarrow, bool forwards=true, Label L[]) {
	link l ;
	l.init(g,arrow,forwards) ;
	l.type(L) ;
	return l ;
}

link link(path g, bool arrow=linkarrow, bool forwards=true ... Label []L) {
	return link(g,arrow,forwards,L) ;
}

/* Link operations */

void type(link l, Label L) {
	l.type(L) ;
}

void type(link l, type t) {
	l.type(t) ;
}

type operator @ (link l, int i) { return l.type[i] ; } 

link copy(link l) { return l.copy() ; } 

link operator * (transform T, link l) {
	link ll = copy(l) ;
	ll.transform(T) ;
	return ll ;
}

link reverse(link orig) {
	type type[] = map(
		new type(type t) { return reverse(t,orig.g) ; },
		orig.type) ;
	bool forwards ;
	if (orig.forwards)
		forwards=false ;
	else
		forwards=true ;
	return link(reverse(orig.g), orig.arrow, forwards, type) ;
}

void test_link (picture pic=currentpicture) { 
	unitsize(pic,2cm) ;
	link l1 = link(W--E..N..cycle,"A") ;
	link l2 = rotate(45)*shift(3W)*scale(2)*l1 ;
	render(pic,l1,l2) ;
}

addtest("test_link",test_link,"link") ;

/**************************************
 *                                    * 
 *            Adjusting               * 
 *                                    * 
 **************************************/

/* Process type lists */ 

typedef type [] typeprocess (type []) ; 

typeprocess Discard = new type [] (type[]) { return new type[] ; } ;

typeprocess Map (type f(type)) { 
	return new type[] (type t[]) { 
		return map(f,t) ; 
	} ; 
} 

typeprocess Reverse (path g) { 
	return Map(new type (type t) { return reverse(t,g) ; }) ;
}

/* Filters */

typedef bool typetest (type) ;

typetest Always = new bool (type) { return true ; } ;
typetest Never = new bool (type) { return false ; } ;

typedef bool inttest (int) ; 

inttest Always = new bool (int) { return true ; } ; 
inttest Never = new bool (int) { return false ; } ; 
inttest Is (int n) { return new bool (int p) { return n==p ; } ; }

typedef bool[] sieve (type []) ;

sieve Test (typetest test) { 
	return new bool [] (type t[]) { 
		bool b[] ;
		for (int i=0; i<t.length; ++i) 
			b.push(test(t[i])) ; 
		return b ; 
	} ; 
}

sieve Testindex (inttest test) { 
	return new bool [] (type t[]) { 
		bool b[] ;
		for (int i=0; i<t.length ; ++i) 
			b.push(test(i)) ; 
		return b ; 
	} ; 
}

sieve All = Test(Always) ;
sieve None = Test(Never) ;
sieve Index (int n) { return Testindex(Is(n)) ; }

/**************************************
 *                                    * 
 *             Merging                * 
 *                                    * 
 **************************************/

/* Join paths 

   keepBegin
  |---------.--------> p
             ⋅
              \
               \
                \
                 ⋅
       |----------`-------> q
                   keepEnd

*/

path join (path begin, path end, 
		position keepBegin, position keepEnd,
		bool revBegin=false, bool revEnd=false ... path [] ps) {
	path p,q ;
	if (revBegin)
		p = reverse(begin) ;
	else 
		p = begin ;
	if (revEnd)
		q = reverse(end) ;
	else 
		q = end ;

	real px,qx ;
	if (keepBegin.relative)
		px = keepBegin.position.x*length(p) ;
	else 
		px = keepBegin.position.x ;
	if (keepEnd.relative)
		qx = length(q)*(1-keepEnd.position.x) ;
	else 
		qx = length(q)-keepEnd.position.x ;

	if (ps.length==0) {
		return subpath(p,0,px){dir(p,px)}..{dir(q,qx)}subpath(q,qx,length(q)) ;
	}

	path mid = ps[0] ;
	for (int i=1;i<ps.length;++i)
		mid=mid..ps[i] ;
	return subpath(p,0,px){dir(p,px)}..mid..{dir(q,qx)}subpath(q,qx,length(q)) ;
}


/* Merge paths */ 

typedef path pathmerge (path,path) ; 

position defaultjoinkeep = .1 ;

pathmerge join (position keepBegin=defaultjoinkeep, 
		position keepEnd=defaultjoinkeep,
		bool revBegin=false, bool revEnd=false 
		... path [] ps) {
	return new path (path p, path q) {
		return join(p,q,keepBegin,keepEnd,revBegin,revEnd...ps) ; 
	} ;
}

pathmerge join = join() ;


/* Merge type lists */ 

typedef type [] typemerge (type [], type []) ;

typemerge Discard = new type [] (type[], type[]) { return new type[] ; } ;
typemerge First = new type [] (type t[], type[]) { return t ; } ;
typemerge Second = new type [] (type [], type t[]) { return t ; } ;
typemerge Concat = 
	new type [] (type t1[], type t2[]) { return concat(t1,t2) ; } ;

/* Composing typeproccess's and typemerge's */

typeprocess compose (typeprocess tp1, typeprocess tp2) {
	return new type [] (type t[]) { return tp2(tp1(t)); } ; 
}

typeprocess operator * (typeprocess tp1, typeprocess tp2) {
	return compose (tp2,tp1) ; 
}

typemerge compose (typemerge tm, typeprocess tp) {
	return new type [] (type t1[], type t2[]) { return tp(tm(t1,t2)); } ; 
}

typemerge operator * (typeprocess tp, typemerge tm) {
	return compose (tm,tp) ; 
}

typemerge compose (typeprocess tp1, typeprocess tp2, typemerge tm) {
	return new type [] (type t1[], type t2[]) { return tm(tp1(t1),tp2(t2)); } ; 
}

struct typeprocess2 { 
	typeprocess tp1 ; 
	typeprocess tp2 ; 
}

typeprocess2 typeprocess2 (typeprocess tp1, typeprocess tp2) { 
	typeprocess2 tps = new typeprocess2 ; 
	tps.tp1 = tp1 ; 
	tps.tp2 = tp2 ; 
	return tps ; 
}

typeprocess2 operator ^ (typeprocess tp1, typeprocess tp2) { 
	return typeprocess2(tp1,tp2) ; 
}

typemerge operator * (typemerge tm, typeprocess2 tp) {
	return compose(tp.tp1,tp.tp2,tm) ;
}

/* Merge links */ 

typedef link linkmerge (link,link) ;

linkmerge Discard = new link (link, link) { return new link ; } ;

linkmerge linkmerge(pathmerge pm, typemerge tm) {
	return new link (link l1, link l2) {
		return link (
			pm(l1.g,l2.g),
			l1.arrow,
			l1.forwards,
			tm(l1.type,l2.type)
		) ; 
	} ;
}

linkmerge Join (typemerge tm=First, position keepBegin=defaultjoinkeep, 
		position keepEnd=defaultjoinkeep, 
		bool revBegin=false, bool revEnd=false ... path [] ps) {
	return linkmerge(join(keepBegin,keepEnd,revBegin,revEnd ... ps),tm) ; 
}

linkmerge Default = Join () ; 
linkmerge KeepFirst = Join (keepBegin=1) ; 

linkmerge JoinSecond = new link (link l1, link l2) {
	return reverse(Default(reverse(l2),reverse(l1))) ; 
} ; 

linkmerge defaultlinkmerge = Default ; 

link merge (link l1, link l2, linkmerge m = defaultlinkmerge) {
	return m(l1,l2) ; 
}

void cut (link l1, link l2, linkmerge m1 = defaultlinkmerge, 
		linkmerge m2=Discard) {
	link ll1 = m1(l1,reverse(l2)) ; 
	link ll2 = m2(l2,reverse(l1)) ; 
	l1.init(ll1) ; 
	l2.init(ll2) ; 
} 

void cutvia (link l1, link l2, typemerge tm=First, 
		explicit position keepBegin=defaultjoinkeep, 
		explicit position keepEnd=defaultjoinkeep,
		bool revBegin=false,
		bool revEnd=false ... path [] ps) {
	linkmerge m = Join(tm,keepBegin,keepEnd,revBegin,revEnd ... ps) ;
	cut(l1,l2,m1=m) ;
} 

void test_merge (picture pic=currentpicture) { 
	unitsize(pic,2cm) ;
	link l1 = link(W--E,"A") ;
	link l2 = shift(E+S)*link(W--E,"B") ; 
	render(pic,l1,l2,merge(l1,l2)) ;
}

addtest("test_merge",test_merge,"link","merge") ;

pair endpoint(path l) {
	return point(l,Relative(1)) ;
}

pair proj (pair z, pair A, pair dir) {
	return z+(dot(A-z,dir)/dot(dir,dir))*dir ;
}

void extend(link l,path p,pathmerge pm=join) {
	l.g=pm(l.g,p) ;
}

void extend(link l,pair A, pair dir, pathmerge pm=join) {
 	pair z=endpoint(l) ;
	pair zz=proj(z,A,dir) ;
	l.g=pm(l.g,(zz-dir)--zz) ;
}

void extend(pair A, pair dir, pathmerge pm=join ... link [] ls) {
	for (int i=0; i<ls.length; ++i) 
		extend(ls[i],A,dir,pm) ; 
}

