import common ;
import tests ;
import type ;
import cell ;

/**************************************
 *                                    * 
 *             Rectangle              * 
 *                                    * 
 **************************************/

/* Basic rectangle structure */

struct rectangle { 
	pair A,B,C,D ;
	pair Ab,Ad,Ba,Bc,Cb,Cd,Dc,Da ;
	path path ;
	real margin ;
	real width ; 
	real height ; 
	real round ;
	pair center ; 
	path base(int n=0) {
		pair b,e ;
		if (n%4==0) {
			b=Ab ;
			e=Ba ;
		} else if (n%4==1) {
			b=Bc ;
			e=Cb ;
		}	else if (n%4==2) {
			b=Cd ;
			e=Dc ;
		}	else {
			b=Da ;
			e=Ad ;
		}
		return (b+margin*unit(e-b))--(e+margin*unit(b-e)) ;
	}
	path base ;
	void init(pair A=this.A, pair B=this.B, 
			real height=this.height, real round=this.round, 
			real margin=this.margin) { 
		this.A = A ;
		this.B = B ;
		this.height = height ; 
		this.round=round ;
		this.margin=margin ;

		C = B+height*unit(rotate(90)*(B-A)) ;
		D = A+height*unit(rotate(90)*(B-A)) ;	

		width = abs(B-A) ; 

		Ab=A+round*unit(B-A) ;
		Ad=A+round*unit(D-A) ;
		Ba=B+round*unit(A-B) ;
		Bc=B+round*unit(C-B) ;
		Cb=C+round*unit(B-C) ;
		Cd=C+round*unit(D-C) ;
		Da=D+round*unit(A-D) ;
		Dc=D+round*unit(C-D) ;

		center=(A+C)/2 ;

		path = Ab---Ba::Bc---Cb::Cd---Dc::Da---Ad::cycle ; 
		path base = base(0) ;
	} ;
	void transform (transform T) {
		init(T*A,T*B,abs(T*C-T*B),round,margin) ;
	}
	rectangle copy () {
		rectangle t = new rectangle ;
		t.init(A,B,height,round,margin) ;
		return t ;
	}
}

path operator cast (rectangle t) { return t.path ; }

rectangle rectangle (pair A, pair B, real height, real round, 
		real margin=0) {
	rectangle t = new rectangle ;
	t.init(A,B,height,round,margin) ;
	return t ;
}

rectangle rectangle (pair A=(0,0), real width, real height=1, real angle=0, 
		real round=rectround, real margin=0) {
	pair B = A+width*dir(angle) ; 
	return rectangle(A,B,height,round,margin) ;
}

/* Triangle operations */

rectangle operator * (transform T, rectangle t) { 
	rectangle tt = t.copy() ; 
	tt.transform(T) ; 
	return tt;  
}

void test_rect (picture pic=currentpicture) {
	rectangle r = rectangle((0,0),(2,0),1,.15) ; 
	label(pic,"A",r.A) ;
	label(pic,"B",r.B) ;
	label(pic,"C",r.C) ;
	label(pic,"D",r.D) ;
	draw(pic,r.path) ;
}

addtest("test_rect: rectangle by corners.",test_rect,"rectangle") ;

/**************************************
 *                                    * 
 *        Rectangular cell            * 
 *                                    * 
 **************************************/

/* Basic generator */

cell rectcell (int main, int after, int num1, int num2, int num3,
		Label L, rectangle rect, pair Lshift) {
	cell c ; 
	label(c.pic,ensuremath(L),rect.center+Lshift) ;
	draw(c.pic,rect.path) ; 
	link baseports [] = ports(rect.base(0),after+main+1) ;
	for (int i=main; i<=main+after; ++i) 
		c.port.push(baseports[i]) ;
	c.port.append(ports(rect.base(1),num1)) ;
	c.port.append(ports(rect.base(2),num2)) ;
	c.port.append(ports(rect.base(3),num3)) ;
	for (int i=0; i<main ; ++i) 
		c.port.push(baseports[i]) ;
	return c ;
}

cell rectcell (pair A=(0,0), real w=1, real h=1,
		Label L, 
		int main=0, int after=0, 
		int num1=0, int num2=0, int num3=0, 
		real round=rectround, real margin=0, pair dir=S, pair Lshift=(0,0)) {
	rectangle r = rotate(degrees(dir)+90,A)*
		rectangle(A,width=w,height=h,angle=0,round=round,margin=margin) ;
	return rectcell(main,after,num1,num2,num3,L,r,Lshift) ;
}

void test_rectcell1 (picture pic=currentpicture) {
	render(pic,rectcell("")) ; 
	render(pic,shift(2E)*rectcell(num2=3,L="$s$")) ; 
	render(pic,shift(2W)*rectcell(L="$s$",dir=S+W)) ; 
	dot(pic,2W) ;
}

addtest("test_rectcell1: basic rectcells, with transforms.",
	test_rectcell1,"rectcell") ;

cell rectcell (pair A=(0,0), real w=1, real h=1,
		Label L, 
		type main=type(Label("")), 
		type before[] = new type[], 
		type after[] = new type[],
		type right[] = new type[],
		type top[] = new type[],
		type left[] = new type[],
		real round=rectround, pair dir=S, pair Lshift=(0,0)) {
	cell c=rectcell(A,w,h,L,before.length,after.length,right.length,
		top.length,left.length,round,dir,Lshift) ;
	type t[]=new type[] {main} ;
	t.append(after) ;
	t.append(right) ;
	t.append(top) ;
	t.append(left) ;
	t.append(before) ;
	for (int i=0; i<t.length; ++i) 
		c.port[i].type(t[i]) ;
	return c ;
}

void test_rectcell2 (picture pic=currentpicture) {
	render(pic,rectcell("s",type("A"))) ; 
	render(pic,shift(2E)*rectcell("t",type("A"),top=types("",ellip,""))) ; 
}

addtest("test_rectcell2: rectcells by types.",
	test_rectcell2,"types","type","rectcell") ;

