import common ;
import tests ;
import cell ;

/**************************************
 *                                    * 
 *             Triangle               * 
 *                                    * 
 **************************************/

/* Basic triangle structure */

struct triangle { 
	pair A,B,C ;
	pair Ab,Ac,Ba,Bc ;
	path path ;
	path base ;
	real round ;
	real base() {	return (length(B-A)) ; }
	real height() {	return (abs(ypart(conj(unit(B-A))*C-A))) ; }
	pair center() { return (A+B+C)/3; }
	void init(pair A=this.A, pair B=this.B, pair C=this.C,
			real round=this.round) { 
		this.A = A ;
		this.B = B ;
		this.C = C ;
		this.round=round ;

		Ab=A+round*unit(B-A) ;
		Ac=A+round*unit(C-A) ;
		Ba=B+round*unit(A-B) ;
		Bc=B+round*unit(C-B) ;

		path = C---Bc::Ba---Ab::Ac---cycle ; 
		base = Ab--Ba ;
	} ;
	void transform (transform T) {
		init(T*A,T*B,T*C,round) ;
	}
	triangle copy () {
		triangle t = new triangle ;
		t.init(A,B,C,round) ;
		return t ;
	}
}

path operator cast (triangle t) { return t.path ; }

triangle triangle (explicit pair A, explicit pair B, explicit pair C, 
		real round) {
	triangle t = new triangle ;
	t.init(A,B,C,round) ;
	return t ;
}

/* Triangle operations */

triangle operator * (transform T, triangle t) { 
	triangle tt = t.copy() ; 
	tt.transform(T) ; 
	return tt;  
}

triangle triangle (real base=defaultcellwidth, real height=defaultcellheight,
		pair dir = S, real round=defaultround) {
	pair A,B,C ;
	A=(-height/2,-base/2) ;
	B=(-height/2,base/2) ;
	C=(height/2,0) ;
	return (rotate(degrees(dir))*triangle(A,B,C,round)) ;
}

void test_triangle (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	label("$O_1$") ;
	triangle t=triangle() ;
	label(pic,Label("$A_1$",position=3.5,align=RightSide),t.path) ;
	label(pic,Label("$B_1$",position=1.5,align=RightSide),t.path) ;
	label(pic,Label("$C_1$",position=0,align=RightSide),t.path) ;
	draw(pic,t.path) ;
	label("$O_2$",(2,0)) ;
	triangle t=shift(2,0) * triangle(1,1,W) ;
	label(pic,Label("$A_2$",position=3.5,align=RightSide),t.path) ;
	label(pic,Label("$B_2$",position=1.5,align=RightSide),t.path) ;
	label(pic,Label("$C_2$",position=0,align=RightSide),t.path) ;
	draw(pic,t.path) ;
}

addtest("test_triangle: triangles by base and height, with transforms.",
	test_triangle,"triangle") ;

/**************************************
 *                                    * 
 *             Tricell                * 
 *                                    * 
 **************************************/

/* Basic generator */

cell tricell (int numAux, Label L, triangle t=triangle(),
		pair dir=S, pair Lshift=(0,0)) {
	picture pic ; 
	label(pic, ensuremath(L), t.center()+Lshift);
	draw(pic,t.path) ;
	cell c ; 
	c.pic = pic ; 
	c.port.push(port(t.C,dir)) ; 
	c.port.append(ports(t.base,numAux,-dir,forwards=false)) ;
	return c ;
}

void test_tricell1 (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	cell cell = tricell(2, Label("c")) ;
	render(pic,cell) ;
	cell = rotate(45)*shift(4E)*cell ;
	render(pic,cell) ;
}

addtest("test_tricell1: tricell by number of auxiliary ports.",
	test_tricell1,"tricell") ;

/* Generate tricell based on labels */

cell tricell (type type[], Label L, triangle t=triangle(),
		pair dir=S, pair Lshift=(0,0)) {
	int numAux = max(0,type.length - 1);
	cell sc = tricell (numAux, L, t, dir, Lshift) ;
	for (int i=0; i<type.length; ++i) 
		sc.port[i].type(type[i]) ;
	return sc ;
}	

cell tricell (Label label[], Label L, triangle t=triangle(),
		pair dir=S, pair Lshift=(0,0), bool reverse=false) {
	int numAux = max(0,label.length - 1);
	type types[] ;
	for (int i=0; i<label.length; ++i) {
		Label l = label[i] ;
		if (i==0 || (numAux>1 && i==1))
			l = Label(l,align=LeftSide) ; 
		else 
			l = Label(l,align=RightSide) ; 
		if (i==0)
			types.push(type(l,forwards=!reverse)) ;
		else
			types.push(type(l,forwards=reverse)) ;
	}
	return tricell(types,L,t,dir,Lshift) ;
}	

cell tricell(Label L, triangle t, bool reverse=false ... Label[] labels) {
	return tricell(labels,L,t,reverse=reverse) ;
}

cell tricell(Label L, bool reverse=false ... Label[] labels) {
	return tricell(labels,reverse=reverse,L) ;
}

void test_tricell2 (picture pic=currentpicture) {
	unitsize(pic,1cm) ;
	cell cell = tricell(new Label[]{"A","B","C"}, Label("\otimes")) ;
	cell = shift(4S)*cell ;
	cell.render(pic) ;
	cell = rotate(45)*cell ;
	render(pic,cell) ;
	cell cell = tricell("t","A","B","","","B") ;
	cell = shift(4E)*rotate(90)*cell ;
	render(pic,cell) ;
}

addtest("test_tricell2: tricell by types.",test_tricell2,"tricell") ;

/* Fake cell to simulate vertical link 
   link, with the same geometry as tricell(1,"",t,dir) */

cell vert (Label pL="", Label aL="", triangle t=triangle(), pair dir=S) {
	cell sc = tricell (new Label[]{pL,aL}, "", t, dir) ; 
	picture pic ;
	path g0 = sc.port[0].g ;
	path g1 = sc.port[1].g ;
	path g = point(g0,0)---point(g1,0) ;
	draw(pic,g) ;
	sc.pic=pic ;
	return sc ;
}


