import common ;
/**************************************
 *                                    * 
 *             Type                   * 
 *                                    * 
 **************************************/

/* Utility */

align reverse(align align) {
	if (align.default) 
		return Relative(W) ;
	if (align.relative)
		return Relative(-align.dir) ;
	return align.copy() ;
}

Label reverse(Label L, path g) {
	position pos =
		L.defaultposition ? Relative(.5) : 
		L.position.relative ? Relative(1-L.position.position.x) :
		length(g)-L.position.position.x ;
	return Label(L,position=pos,align=reverse(L.align)) ;
}

string ensuremath(string s) {
	return "\ensuremath{"+s+"}" ;
}

Label ensuremath(Label L) {
	Label LL = L.copy() ;
	LL.s = ensuremath(LL.s) ;
	return LL ;
}

pair point(path g, position pos) {
	real t = pos.relative ? reltime(g,pos.position.x) : pos.position.x ;
	return point(g,t) ;
}

pair dir(path g, position pos) {
	real t = pos.relative ? reltime(g,pos.position.x) : pos.position.x ;
	return dir(g,t) ;
}

real degrees(path g, position pos) {
	return degrees(dir(g,pos)) ;
}

path subpath(path g, position a, position b) {
	real ta = a.relative ? reltime(g,a.position.x) : a.position.x ;
	real tb = b.relative ? reltime(g,b.position.x) : b.position.x ;
	return subpath(g,ta,tb) ;
}

/* Dirty hack to change rendering of links depending on special type labels */

string mute   = "" ;

// render $\cdots$ at right angle with link path
string ellipsis  = "@ellipsis@" ; 

// don't render link path
string kill		= "@kill@" ;	

string ellip  = ellipsis+kill ;

// unrendered fake type
string none		= "@none@" ; 

bool hastag(string s,string t) {
	return find(s,t) >= 0 ;
}
bool hastag(Label L,string t) {
	return hastag(L.s,t) ;
}

/* Basic type structure */

struct type {
	Label L ;
	bool arrow ; 
	bool forwards ;
	void init(Label L, bool arrow, bool forwards) {
		this.L = L ;
		this.arrow = arrow ;
		this.forwards = forwards ;
	}
	type copy() {
		type t = new type ;
		t.init(L.copy(),arrow,forwards) ;
		return t ;
	}
	bool do_out() { return ! hastag(L,none) ; } 
	
	void out(picture pic=currentpicture, path g) {
		Label LL = forwards ? L : reverse(L,g) ;
		LL=ensuremath(LL) ;
		position pos=LL.defaultposition ? Relative(.5) : LL.position ;
		path gg =  forwards ? g : reverse(g) ;
		if (hastag (L,ellipsis))
			label(pic,
				rotate(degrees(dir(g,pos))+90)*
				Label("$\cdots$",align=Center,position=pos),
				g) ;
		else if (!hastag(L,none) && !hastag(L,kill)) {
			LL.out(pic,gg) ;
			if (arrow)
				add(pic,
					arrow(gg, 
						size=2mm, 
						position=pos,
						center=true
					)
				) ;
		}
	}
	void transform(transform T) {
		L.transform(T) ;
	}
}

type operator init() { return new type ; } 

type type(explicit Label L, bool arrow=linkarrow, bool forwards=true) {
	type t ;
	t.init(L,arrow,forwards) ;
	return t ;
}

type type(string L, bool arrow=linkarrow, bool forwards=true, 
		align align=RightSide) {
	return type(Label(L,align=align),arrow,forwards) ; 
}

type tkill =type(kill) ;
type tellip=type(ellip) ;
type tnone =type(none) ;
type tmute =type(mute) ;
type rmute =type(forwards=false,mute) ;


type [] types(Label L[], bool arrow=linkarrow, bool forwards=true) {
	type t[] ;
	for (int i=0; i<L.length; ++i)
		t.push(type(L[i],arrow,forwards)) ;
	return t ;
}

type [] types(string s[], bool arrow=linkarrow, bool forwards=true,
		align align=RightSide, bool fix_last_align=true) {
	Label L[] ;
	for (int i=0; i<s.length; ++i)
		L.push(Label(s[i],align=align));
	if (fix_last_align && s.length>0) 
		L[s.length-1].align.dir=-align.dir ;
	return types(L,arrow,forwards) ;
}

type [] types(bool arrow=linkarrow, bool forwards=true ... Label [] L) {
	return types(L,arrow,forwards) ; 
}

type [] types(bool arrow=linkarrow, bool forwards=true, 
		align align=RightSide, bool fix_last_align=true ... string [] s) {
	return types(s,arrow,forwards,align,fix_last_align) ;
}

/* Type operations */

type copy(type t) { return t.copy() ; } 

type unforwards(type t) {
	bool forwards ;
	if (t.forwards)
		forwards=false ;
	else
		forwards=true ;
	return type(t.L,t.arrow,forwards) ;
}

type [] unforwards(type t[]) {
	return map(unforwards,t) ;
}

type [][] unforwards(type t[][]) {
	type u[][] ;
	for (int i=0 ; i<t.length ; ++i)
		u.push(unforwards(t[i])) ;
	return u ;
}

type reverse(type t, path g) {
	bool forwards ;
	if (t.forwards)
		forwards=false ;
	else
		forwards=true ;
	return type(reverse(t.L,g),t.arrow,forwards) ;
}

type [] reverse(type t[], path g) {
	return map(new type(type ty) { return reverse(ty,g) ; }, t) ;
}

